#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['scgen', 'scgen.models']

package_data = \
{'': ['*']}

install_requires = \
['future-fstrings',
 'get_version',
 'anndata',
 'tensorflow',
 'numpy',
 'scipy',
 'pandas',
 'matplotlib']

extras_require = \
{'test': ['pytest']}

setup(name='scgen',
      version='1.0.0',
      description='ScGen - Predicting single cell perturbations',
      author='Mohammad lotfollahi',
      author_email='mohammad.lotfollahi@helmholtz-muenchen.de',
      url='https://github.com/theislab/scgen',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
